/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** Contact: Qt Software Information (qt-info@nokia.com)
**
** This file is part of a Qt Solutions component.
**
** GNU General Public License Usage
** This file may be used under the terms of the GNU
** General Public License versions 2.0 or 3.0 as published by the Free
** Software Foundation and appearing in the file LICENSE.GPL included in
** the packaging of this file.  Please review the following information
** to ensure GNU General Public Licensing requirements will be met:
** http://www.fsf.org/licensing/licenses/info/GPLv2.html and
** http://www.gnu.org/copyleft/gpl.html.  In addition, as a special
** exception, Nokia gives you certain additional rights. These rights
** are described in the Nokia Qt GPL Exception version 1.3, included in
** the file GPL_EXCEPTION.txt in this package.
**
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
**
****************************************************************************/
#include "lifecycle.h"
#include "stickman.h"
#include "node.h"
#include "animation.h"
#include "graphicsview.h"
#ifdef QT_STATEMACHINE_SOLUTION
# include "qsignaltransition.h"
# include "qeventtransition.h"
# include "qpropertyanimation.h"
# include "qanimationstate.h"
# include "qparallelanimationgroup.h"
# include "qstatemachine.h"
# include "qsignalevent.h"
#endif

#include <QtCore>
#include <QtGui>

class KeyPressTransition: public QtSignalTransition
{
public:
    KeyPressTransition(GraphicsView *receiver, Qt::Key key)
        : QtSignalTransition(receiver, SIGNAL(keyPressed(int))), m_key(key)
    {
    }

    virtual bool testCondition(QEvent *e) const
    {
        if (QtSignalTransition::testCondition(e)) {
            QVariant key = static_cast<QtSignalEvent*>(e)->arguments().at(0);
            return (key.toInt() == int(m_key));
        } 

        return false;
    }
private:
    Qt::Key m_key;
};

class LightningStrikesTransition: public QtEventTransition
{
public:
    LightningStrikesTransition() : QtEventTransition(this, QEvent::Timer) 
    {
        qsrand((uint)QDateTime::currentDateTime().toTime_t());
        startTimer(1000);
    }

    virtual bool testCondition(QEvent *e) const
    {
        return QtEventTransition::testCondition(e) && ((qrand() % 10) == 0);
    }
};

LifeCycle::LifeCycle(StickMan *stickMan, GraphicsView *keyReceiver)
    : m_keyReceiver(keyReceiver), m_stickMan(stickMan)
{
    // Create animation group to be used for all transitions
    m_animationGroup = new QtParallelAnimationGroup();
    const int stickManNodeCount = m_stickMan->nodeCount();
    for (int i=0; i<stickManNodeCount; ++i) {
        QtPropertyAnimation *pa = new QtPropertyAnimation(m_stickMan->node(i), "position");
        m_animationGroup->add(pa);    
    }

    // Set up intial state graph
    m_machine = new QtStateMachine();
    m_machine->setGlobalRestorePolicy(QtState::RestoreProperties);

    m_alive = new QtState(m_machine->rootState());
    m_alive->setObjectName("alive");
    
    // Make it blink when lightning strikes before entering dead animation
    QtState *lightningBlink = new QtState(m_machine->rootState());    
    lightningBlink->setPropertyOnEntry(m_stickMan->scene(), "backgroundBrush", Qt::white);
    lightningBlink->setPropertyOnEntry(m_stickMan, "penColor", Qt::black);
    lightningBlink->setPropertyOnEntry(m_stickMan, "fillColor", Qt::white);
  
    m_dead = new QtState(m_machine->rootState());
    m_dead->setPropertyOnEntry(m_stickMan->scene(), "backgroundBrush", Qt::black);
    m_dead->setPropertyOnEntry(m_stickMan, "penColor", Qt::white);
    m_dead->setPropertyOnEntry(m_stickMan, "fillColor", Qt::black);
    m_dead->setObjectName("dead");
           
    // Idle state (sets no properties)
    m_idle = new QtState(m_alive);
    m_idle->setObjectName("idle");
    m_alive->setInitialState(m_idle);

    // Lightning strikes at random
    m_alive->addTransition(new LightningStrikesTransition(), lightningBlink);
    connectByAnimation(m_machine->rootState(), lightningBlink, m_dead);

    m_machine->setInitialState(m_alive);
}

void LifeCycle::setResetKey(Qt::Key resetKey)
{
    // When resetKey is pressed, enter the idle state and do a restoration animation
    // (requires no animation pointer, since no property is being set in the idle state)
    m_alive->addAnimatedTransition(new KeyPressTransition(m_keyReceiver, resetKey), m_idle);
}

void LifeCycle::setDeathAnimation(const QString &fileName)
{
    QtState *deathAnimation = makeState(m_dead, fileName);
    m_dead->setInitialState(deathAnimation);
}

void LifeCycle::start()
{
    m_machine->start();
}

void LifeCycle::connectByAnimation(QtAbstractStateGroup *parentState, 
                                   QtAbstractStateGroup *s1, QtAbstractState *s2, 
                                   QtAbstractTransition *transition)
{
    QtAnimationState *animationState = new QtAnimationState(m_animationGroup, parentState);
    
    if (transition == 0)
        s1->addTransition(animationState);
    else
        s1->addTransition(transition, animationState);    

    animationState->addFinishedTransition(s2);
}

void LifeCycle::addActivity(const QString &fileName, Qt::Key key)
{
    QtState *state = makeState(m_alive, fileName);
    connectByAnimation(m_alive, m_alive, state, new KeyPressTransition(m_keyReceiver, key));
}

QtState *LifeCycle::makeState(QtAbstractStateGroup *parentState, const QString &animationFileName)
{
    QtState *topLevel = new QtState(parentState);
    
    Animation animation;
    {
        QFile file(animationFileName);
        if (file.open(QIODevice::ReadOnly))
            animation.load(&file);
    }

    const int frameCount = animation.totalFrames();
    QtState *previousState = 0;
    for (int i=0; i<frameCount; ++i) {
        QtState *frameState = new QtState(topLevel);
        
        animation.setCurrentFrame(i);
        const int nodeCount = animation.nodeCount();
        for (int j=0; j<nodeCount; ++j)
            frameState->setPropertyOnEntry(m_stickMan->node(j), "position", animation.nodePos(j));

        if (previousState == 0)
            topLevel->setInitialState(frameState);
        else
            connectByAnimation(topLevel, previousState, frameState);
        previousState = frameState;
    }

    // Loop
    connectByAnimation(topLevel, previousState, topLevel->initialState());

    return topLevel;

}

LifeCycle::~LifeCycle()
{
    delete m_machine;
    delete m_animationGroup;
}
