/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** Contact: Qt Software Information (qt-info@nokia.com)
**
** This file is part of a Qt Solutions component.
**
** GNU General Public License Usage
** This file may be used under the terms of the GNU
** General Public License versions 2.0 or 3.0 as published by the Free
** Software Foundation and appearing in the file LICENSE.GPL included in
** the packaging of this file.  Please review the following information
** to ensure GNU General Public Licensing requirements will be met:
** http://www.fsf.org/licensing/licenses/info/GPLv2.html and
** http://www.gnu.org/copyleft/gpl.html.  In addition, as a special
** exception, Nokia gives you certain additional rights. These rights
** are described in the Nokia Qt GPL Exception version 1.3, included in
** the file GPL_EXCEPTION.txt in this package.
**
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
**
****************************************************************************/


#include "qanimation.h"
#include "qanimation_p.h"

#include <QtCore/QRect>
#include <QtCore/QLineF>
#include <QtCore/QLine>
#include <QtCore/QReadWriteLock>
#include <QtCore/qdebug.h>

QT_BEGIN_NAMESPACE

/*!
    \class QtAnimation
    
    \brief The QtAnimation class provides an abstract base class for animations.
    
    \preliminary

    This class is part of {The Animation Framework}. It serves as a base class
    for property and item animations, with functions for shared functionality.

    If you want to create an animation, you should look at QtPropertyAnimation instead.

    You can then set start and end values for the property by calling
    setStartValue() and setEndValue(), and finally call start() to
    start the animation. When control goes back to the event loop, QtAnimation
    will interpolate the property of the target object and emit the valueChanged
    signal. To react to a change in the current value you have to reimplement the
    updateCurrentValue virtual method.

    There are two ways to affect how QtAnimation interpolates the values.
    You can set an easing curve by calling setEasingCurve(), and configure the
    duration by calling setDuration(). You can change how the QVariants are
    interpolated by creating a subclass of QtAnimation, and reimplementing the
    virtual interpolated() function.


    \sa QtPropertyAnimation, {The Animation Framework}
*/

/*!
    \fn void QtAnimation::valueChanged(const QVariant &value)

    QtAnimation emits this signal whenever the current \a value changes.

    \sa currentValue, startValue, endValue
*/

/*!
    \fn void QtAnimation::durationChanged(int duration)

    QtAnimation emits this signal whenever the \a duration of the animation
    changes.

    \sa setDuration(), duration()
*/

static bool animationValueLessThan(const QtAnimation::KeyFrame &p1, const QtAnimation::KeyFrame &p2)
{
    return p1.first < p2.first;
}

template<> Q_INLINE_TEMPLATE QRect _q_interpolate(const QRect &f, const QRect &t, qreal progress)
{
    QRect ret;
    ret.setCoords(_q_interpolate(f.left(), t.left(), progress),
                  _q_interpolate(f.top(), t.top(), progress),
                  _q_interpolate(f.right(), t.right(), progress),
                  _q_interpolate(f.bottom(), t.bottom(), progress));
    return ret;
}

template<> Q_INLINE_TEMPLATE QRectF _q_interpolate(const QRectF &f, const QRectF &t, qreal progress)
{
    qreal x1, y1, w1, h1;
    f.getRect(&x1, &y1, &w1, &h1);
    qreal x2, y2, w2, h2;
    t.getRect(&x2, &y2, &w2, &h2);
    return QRectF( _q_interpolate(x1, x2, progress), _q_interpolate(y1, y2, progress),
                  _q_interpolate(w1, w2, progress), _q_interpolate(h1, h2, progress));
}

template<> Q_INLINE_TEMPLATE QLine _q_interpolate(const QLine &f, const QLine &t, qreal progress)
{
    return QLine( _q_interpolate(f.p1(), t.p1(), progress), _q_interpolate(f.p2(), t.p2(), progress));
}

template<> Q_INLINE_TEMPLATE QLineF _q_interpolate(const QLineF &f, const QLineF &t, qreal progress)
{
    return QLineF( _q_interpolate(f.p1(), t.p1(), progress), _q_interpolate(f.p2(), t.p2(), progress));
}

void QtAnimationPrivate::convertValues(int t)
{
    //this ensures that all the keyFrames are of type t
    for (int i = 0; i < keyFrames.count(); ++i) {
        QtAnimation::KeyFrame &pair = keyFrames[i];
        if (pair.second.userType() != t)
            pair.second.convert(static_cast<QVariant::Type>(t));
    }
}

QtAnimationPrivate *QtAnimationPrivate::get(QtAnimation *q)
{
    return q->d_func();
}

/*!
    \fn void QtAnimation::updateCurrentValue(const QVariant &value) = 0;
    This pure virtual function is called when the animated value is changed.
    \a value is the new value.
*/

void QtAnimationPrivate::updateCurrentValue()
{
    Q_Q(QtAnimation);
    const qreal progress = easing.valueForProgress(((duration == 0) ? qreal(1) : qreal(currentTime) / qreal(duration)));

    if (progress < currentInterval.start.first || progress > currentInterval.end.first) {
        //let's update currentInterval
        QtAnimation::KeyFrames::const_iterator itStart = qLowerBound(keyFrames.constBegin(), keyFrames.constEnd(), qMakePair(progress, QVariant()), animationValueLessThan);
        QtAnimation::KeyFrames::const_iterator itEnd = itStart;

        // If we are at the end we should continue to use the last keyFrames in case of extrapolation (progress > 1.0).
        // This is because the easing function can return a value slightly outside the range [0, 1]
        if (itStart != keyFrames.constEnd()) {

            //this can't happen because we always prepend the default start value there
            if (itStart == keyFrames.begin()) {
                ++itEnd;
                if (itEnd == keyFrames.constEnd())
                    return; //there is no upper bound
            } else {
                --itStart;
            }

            //update all the values of the currentInterval
            currentInterval.start = *itStart;
            currentInterval.end = *itEnd;
        }
    }

    const qreal startProgress = currentInterval.start.first,
                endProgress = currentInterval.end.first;
    const qreal localProgress = (progress - startProgress) / (endProgress - startProgress);
    
    bool changed = false;
    {
      //we do that here in a limited scope so that ret is dereferenced and frees memory
      //and the call to updateCurrentValue can recreate QVariant of the same type (for ex. in
      //QGraphicsItem::setPos
      QVariant ret = q->interpolated(currentInterval.start.second,
                                    currentInterval.end.second,
                                    localProgress);
      if (currentValue != ret) {
          changed = true;
          qSwap(currentValue, ret);
      }
    }

    if (changed) {
        //the value has changed
        q->updateCurrentValue(currentValue);
#ifndef QT_EXPERIMENTAL_SOLUTION
        if (connectedSignals & changedSignalMask)
#endif
            emit q->valueChanged(currentValue);
    }
}


QVariant QtAnimationPrivate::valueAt(qreal step) const
{
    QtAnimation::KeyFrames::const_iterator result = 
        qBinaryFind(keyFrames.begin(), keyFrames.end(), qMakePair(step, QVariant()), animationValueLessThan);
    if (result != keyFrames.constEnd())
        return result->second;

    return QVariant();
}


void QtAnimationPrivate::setValueAt(qreal step, const QVariant &value)
{
    if (step < qreal(0.0) || step > qreal(1.0)) {
        qWarning("QtAnimation::setValueAt: invalid step = %f", step);
        return;
    }

    QtAnimation::KeyFrame pair(step, value);

    QtAnimation::KeyFrames::iterator result = qLowerBound(keyFrames.begin(), keyFrames.end(), pair, animationValueLessThan);
    if (result == keyFrames.end() || result->first != step) {
        keyFrames.insert(result, pair);
    } else {
        if (value.isValid())
            result->second = value; //remove the previous value
        else if (step == 0 && !hasStartValue && defaultStartValue.isValid())
            result->second = defaultStartValue; //we reset to the default start value
        else
            keyFrames.erase(result); //replace the previous value
    }

    currentInterval.start.first = 2; // this will force the refresh
    updateCurrentValue();
}

/*!
    Construct a QtAnimation object. \a parent is passed to QtAbstractAnimation's
    constructor.
*/
QtAnimation::QtAnimation(QObject *parent) : QtAbstractAnimation(*new QtAnimationPrivate, parent)
{
   d_func()->init();
}

/*!
    \internal
*/
QtAnimation::QtAnimation(QtAnimationPrivate &dd, QObject *parent) : QtAbstractAnimation(dd, parent)
{
   d_func()->init();
}

/*!
    \property QtAnimation::easingCurve
    \brief the easing curve of the animation

    This property defines the easing curve of the animation. By default, a
    linear easing curve is used, resulting in linear interpolation of the
    end property. For many animations, it's useful to try different easing
    curves, including QtEasingCurve::InCirc, which provides a circular entry curve,
    and QtEasingCurve::InOutElastic, which provides an elastic effect on the values
    of the interpolated property.

    The easing curve is used with the interpolator, the interpolated() virtual
    function, the animation's duration, and loopCount, to control how the
    current value changes as the animation progresses.
*/
QtEasingCurve QtAnimation::easingCurve() const
{
    Q_D(const QtAnimation);
    return d->easing;
}

void QtAnimation::setEasingCurve(const QtEasingCurve &easing)
{
    Q_D(QtAnimation);
    d->easing = easing;
    d->updateCurrentValue();
}

Q_GLOBAL_STATIC(QVector<QtAnimation::Interpolator>, registeredInterpolators)
Q_GLOBAL_STATIC(QReadWriteLock, registeredInterpolatorsLock)

/*!
    \fn void qRegisterAnimationInterpolator(QtAnimation::Interpolator func)
    \relates QtAnimation
    \threadsafe

    Registers a custom interpolator \a func for the template type \c{T}.
    The interpolator has to be registered before the animation is constructed.
    To unregister (and use the default interpolator) set \a func to 0.
 */

/*!    
    \typedef QtAnimation::Interpolator
    
    This is a typedef for a pointer to a function with the following
    signature:
    \code
    QVariant myInterpolator(const QVariant &from, const QVariant &to, qreal progress);
    \endcode
    
*/

/*! \internal
 * Registers a custom interpolator \a func for the specific \a interpolationType.
 * The interpolator has to be registered before the animation is constructed.
 * To unregister (and use the default interpolator) set \a func to 0. ###
 */
void QtAnimation::registerInterpolator(QtAnimation::Interpolator func, int interpolationType)
{
    // will override any existing interpolators
    QWriteLocker locker(registeredInterpolatorsLock());
    if (int(interpolationType) >= registeredInterpolators()->count())
        registeredInterpolators()->resize(int(interpolationType) + 1);
    registeredInterpolators()->replace(interpolationType, func);
}

static QtAnimation::Interpolator getInterpolator(int interpolationType)
{
    QReadLocker locker(registeredInterpolatorsLock());
    QtAnimation::Interpolator ret = 0;
    if (interpolationType < registeredInterpolators()->count()) {
        ret = registeredInterpolators()->at(interpolationType);
        if (ret) return ret;
    }

    switch(interpolationType)
    {
    case QMetaType::Int:
        return _q_interpolateVariant<int>;
    case QMetaType::Double:
        return _q_interpolateVariant<double>;
    case QMetaType::Float:
        return _q_interpolateVariant<float>;
    case QMetaType::QLine:
        return _q_interpolateVariant<QLine>;
    case QMetaType::QLineF:
        return _q_interpolateVariant<QLineF>;
    case QMetaType::QPoint:
        return _q_interpolateVariant<QPoint>;
    case QMetaType::QPointF:
        return _q_interpolateVariant<QPointF>;
    case QMetaType::QSize:
        return _q_interpolateVariant<QSize>;
    case QMetaType::QSizeF:
        return _q_interpolateVariant<QSizeF>;
    case QMetaType::QRect:
        return _q_interpolateVariant<QRect>;
    case QMetaType::QRectF:
        return _q_interpolateVariant<QRectF>;
    default:
        return 0; //this type is not handled
    }
}

/*!
    \property QtAnimation::duration
    \brief the duration of the animation

    This property describes the duration of the animation. The default
    duration is 250 milliseconds.

    \sa QtAbstractAnimation::duration()
 */
int QtAnimation::duration() const
{
    Q_D(const QtAnimation);
    return d->duration;
}

void QtAnimation::setDuration(int msecs)
{
    Q_D(QtAnimation);
    if (msecs < 0) {
        qWarning("QtAnimation::setDuration: cannot set a negative duration");
        return;
    }
    if (d->duration == msecs)
        return;
    d->duration = msecs;
    d->updateCurrentValue();
    emit durationChanged(msecs);
}

/*!
    \property QtAnimation::startValue
    \brief the optional start value of the animation

    This property describes the optional start value of the animation. If
    omitted, or if a null QVariant is assigned as the start value, the
    animation will use the current position of the end when the animation
    is started.

    \sa endValue
*/
QVariant QtAnimation::startValue() const
{
    return valueAt(0);
}

void QtAnimation::setStartValue(const QVariant &value)
{
    setValueAt(0, value);
}

/*!
    \property QtAnimation::endValue
    \brief the end value of the animation

    This property describes the end value of the animation.

    \sa startValue
 */
QVariant QtAnimation::endValue() const
{
    return valueAt(1);
}

void QtAnimation::setEndValue(const QVariant &value)
{
    setValueAt(1, value);
}


/*!
    Returns the key frame value for the given \a step. The given \a step
    must be in the range 0 to 1. If there is no keyframe for \a step,
    it returns an invalid QVariant.

    \sa keyFrames(), setValueAt()
*/
QVariant QtAnimation::valueAt(qreal step) const
{
    Q_D(const QtAnimation);
    if (step == 0 && !d->hasStartValue)
        return QVariant(); //special case where we don't have an explicit startValue

    return d->valueAt(step);
}

/*!    
    \typedef QtAnimation::KeyFrame

    This is a typedef for QPair<qreal, QVariant>.
*/
/*!    
    \typedef QtAnimation::KeyFrames

    This is a typedef for QVector<KeyFrame>
*/

/*!
    Creates a key frame at the given \a step with the given \a value.
    The given \a step must be in the range 0 to 1.

    \sa setKeyFrames(), valueAt()
*/

void QtAnimation::setValueAt(qreal step, const QVariant &value)
{
    Q_D(QtAnimation);
    if (step == 0)
        d->hasStartValue = value.isValid();
    d->setValueAt(step, value);
}

/*!
    Returns the key frames of this animation.

    \sa valueAt()
*/

QtAnimation::KeyFrames QtAnimation::keyFrames() const
{
    Q_D(const QtAnimation);
    QtAnimation::KeyFrames ret = d->keyFrames;
    //in case we added the default start value, we remove it
    if (!d->hasStartValue && !ret.isEmpty() && ret.at(0).first == 0)
        ret.remove(0);
    return ret;
}

/*!
    Replaces the current set of key frames with the given \a keyFrames.
    the step of the key frames must be in the range 0 to 1.

    \sa keyFrames(), valueAt()
*/

void QtAnimation::setKeyFrames(const KeyFrames &keyFrames)
{
    Q_D(QtAnimation);
    d->keyFrames = keyFrames;
    qSort(d->keyFrames.begin(), d->keyFrames.end(), animationValueLessThan);
    d->currentInterval.start.first = 2; // this will force the refresh
    d->hasStartValue = !d->keyFrames.isEmpty() && d->keyFrames.at(0).first == 0;
}

/*!
    \property QtAnimation::currentValue
    \brief the current value of the animation

    This property describes the current value; an interpolation between the
    start value and the end value, using the current time for progress.

    QtAnimation calls the virtual updateCurrentValue() function when the
    current value changes. This is particularily useful for subclasses that
    need to track updates.

    \sa startValue, endValue
 */
QVariant QtAnimation::currentValue() const
{
    Q_D(const QtAnimation);
    if (!d->currentValue.isValid())
        const_cast<QtAnimationPrivate*>(d)->updateCurrentValue();
    return d->currentValue;
}

/*!
    \reimp
 */
bool QtAnimation::event(QEvent *event)
{
    return QtAbstractAnimation::event(event);
}

/*!
    \reimp
*/
void QtAnimation::updateState(QtAbstractAnimation::State oldState,
                             QtAbstractAnimation::State newState)
{
    Q_UNUSED(oldState);
    Q_UNUSED(newState);
    Q_D(QtAnimation);
    d->currentValue = QVariant(); // this will force the refresh
}



/*!
    This virtual function returns the linear interpolation between variants \a
    from and \a to, at \a progress, usually a value between 0 and 1. You can reimplement
    this function in a subclass of QtAnimation to provide your own interpolation
    algorithm. Note that in order for the interpolation to work with a QtEasingCurve
    that return a value smaller than 0 or larger than 1 (such as QtEasingCurve::InBack)
    you should make sure that it can extrapolate. If the semantic of the datatype
    does not allow extrapolation this function should handle that gracefully.

    \sa QtEasingCurve
 */
QVariant QtAnimation::interpolated(const QVariant &from, const QVariant &to, qreal progress) const
{
    Q_D(const QtAnimation);
    if (d->interpolator == 0) {
        d->interpolator = getInterpolator(from.userType());
        if (d->interpolator == 0) //no interpolator found
            return QVariant();
    }

    return d->interpolator(from, to, progress);
}

/*!
    \reimp
 */
void QtAnimation::updateCurrentTime(int msecs)
{
    Q_D(QtAnimation);
    Q_UNUSED(msecs);
    d->updateCurrentValue();
}

/*!
    gets the default start value that will be used in case no start value is 
    explicitly set.

    \sa setDefaultStartValue(), startValue
*/

QVariant QtAnimation::defaultStartValue() const
{
    Q_D(const QtAnimation);
    return d->defaultStartValue;
}

/*!
    Sets the default start value to \a value that will be used by the animation in case it 
    has no defined start value. This is used by subclasses of QtAnimation to define
    a default behaviour and allow to create an animation with simply a target value.

    \sa defaultStartValue(), startValue
*/

void QtAnimation::setDefaultStartValue(const QVariant &value)
{
    Q_D(QtAnimation);
    d->defaultStartValue = value;
    if (!d->hasStartValue)
      d->setValueAt(0, value);
}

QT_END_NAMESPACE

#include "moc_qanimation.cpp"
