/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** Contact: Qt Software Information (qt-info@nokia.com)
**
** This file is part of a Qt Solutions component.
**
** GNU General Public License Usage
** This file may be used under the terms of the GNU
** General Public License versions 2.0 or 3.0 as published by the Free
** Software Foundation and appearing in the file LICENSE.GPL included in
** the packaging of this file.  Please review the following information
** to ensure GNU General Public Licensing requirements will be met:
** http://www.fsf.org/licensing/licenses/info/GPLv2.html and
** http://www.gnu.org/copyleft/gpl.html.  In addition, as a special
** exception, Nokia gives you certain additional rights. These rights
** are described in the Nokia Qt GPL Exception version 1.3, included in
** the file GPL_EXCEPTION.txt in this package.
**
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
**
****************************************************************************/


#include "qsignaltransition.h"
#include "qsignaltransition_p.h"
#include "qsignalevent.h"
#include <qdebug.h>

QT_BEGIN_NAMESPACE

/*!
  \class QtSignalTransition

  \brief The QtSignalTransition class provides a transition based on a Qt signal.

  

  Typically you would use the overload of QtAbstractState::addTransition()
  that takes a sender and signal as arguments, rather than creating
  QtSignalTransition objects directly. QtSignalTransition is part of \l{The
  State Machine Framework}.

  You can subclass QtSignalTransition and reimplement testCondition() to make a
  signal transition conditional.
*/

QtSignalTransitionPrivate::QtSignalTransitionPrivate()
{
}

QtSignalTransitionPrivate *QtSignalTransitionPrivate::get(QtSignalTransition *q)
{
    return q->d_func();
}

/*!
  Constructs a new signal transition associated with the given \a signal of
  the given \a sender.

  \sa QtAbstractStateGroup::addTransition()
*/
QtSignalTransition::QtSignalTransition(const QObject *sender, const char *signal)
    : QtTransition(*new QtSignalTransitionPrivate,
#ifndef QT_STATEMACHINE_SOLUTION
                  QEvent::Signal
#else
                  QEvent::Type(QEvent::User-1)
#endif
        )
{
    Q_D(QtSignalTransition);
    d->sender = sender;
    d->signal = signal;
    d->signalIndex = -1;
}

/*!
  Destroys this signal transition.
*/
QtSignalTransition::~QtSignalTransition()
{
}

/*!
  \reimp

  The \a event is a QtSignalEvent object.  The default implementation returns
  true if the event's sender and signal index match this transition, and
  returns false otherwise.
*/
bool QtSignalTransition::testCondition(QEvent *event) const
{
    Q_D(const QtSignalTransition);
    QtSignalEvent *se = static_cast<QtSignalEvent*>(event);
    Q_ASSERT(d->signalIndex != -1);
    return (se->sender() == d->sender)
        && (se->signalIndex() == d->signalIndex);
}

/*!
  \reimp
*/
void QtSignalTransition::onTransition()
{
}

QT_END_NAMESPACE
