/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** Contact: Qt Software Information (qt-info@nokia.com)
**
** This file is part of a Qt Solutions component.
**
** GNU General Public License Usage
** This file may be used under the terms of the GNU
** General Public License versions 2.0 or 3.0 as published by the Free
** Software Foundation and appearing in the file LICENSE.GPL included in
** the packaging of this file.  Please review the following information
** to ensure GNU General Public Licensing requirements will be met:
** http://www.fsf.org/licensing/licenses/info/GPLv2.html and
** http://www.gnu.org/copyleft/gpl.html.  In addition, as a special
** exception, Nokia gives you certain additional rights. These rights
** are described in the Nokia Qt GPL Exception version 1.3, included in
** the file GPL_EXCEPTION.txt in this package.
**
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
**
****************************************************************************/


#include "qstate.h"
#include "qstate_p.h"
#include "qsignaltransition.h"

QT_BEGIN_NAMESPACE

/*!
  \class QtState

  \brief The QtState class provides a default state implementation for QtStateMachine.

  

  QtState is part of \l{The State Machine Framework}.

  The setInitialState() function sets the initial state; this is only relevant
  for states that have child states. The child states are mutually exclusive
  states, and the state machine needs to know which child state to enter when
  the parent state is the target of a transition.
*/

QtStatePrivate::QtStatePrivate()
    : initialState(0)
{
}

QtStatePrivate::~QtStatePrivate()
{
}

QtStatePrivate *QtStatePrivate::get(QtState *q)
{
    return q->d_func();
}

const QtStatePrivate *QtStatePrivate::get(const QtState *q)
{
    return q->d_func();
}

/*!
  Constructs a new state with the given \a parent state.
*/
QtState::QtState(QtAbstractStateGroup *parent)
    : QtAbstractStateGroup(*new QtStatePrivate, parent)
{
}

/*!
  \internal
*/
QtState::QtState(QtStatePrivate &dd, QtAbstractStateGroup *parent)
    : QtAbstractStateGroup(dd, parent)
{
}

/*!
  Destroys this state.
*/
QtState::~QtState()
{
}

/*!
  Returns this state's initial state, or 0 if the state has no initial state.
*/
QtAbstractState *QtState::initialState() const
{
    Q_D(const QtState);
    return d->initialState;
}

/*!
  Sets this state's initial state to be the given \a state.
  \a state has to be a child of this state.
*/
void QtState::setInitialState(QtAbstractState *state)
{
    Q_D(QtState);
    if (state && (state->parentState() != this)) {
        qWarning("QtState::setInitialState: state %p is not a child of this state (%p)",
                 state, this);
        return;
    }
    d->initialState = state;
}

/*!
  \reimp
*/
void QtState::onEntry()
{
    QtAbstractStateGroup::onEntry();
}

/*!
  \reimp
*/
void QtState::onExit()
{
    QtAbstractStateGroup::onExit();
}

QT_END_NAMESPACE
