/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** Contact: Qt Software Information (qt-info@nokia.com)
**
** This file is part of a Qt Solutions component.
**
** GNU General Public License Usage
** This file may be used under the terms of the GNU
** General Public License versions 2.0 or 3.0 as published by the Free
** Software Foundation and appearing in the file LICENSE.GPL included in
** the packaging of this file.  Please review the following information
** to ensure GNU General Public Licensing requirements will be met:
** http://www.fsf.org/licensing/licenses/info/GPLv2.html and
** http://www.gnu.org/copyleft/gpl.html.  In addition, as a special
** exception, Nokia gives you certain additional rights. These rights
** are described in the Nokia Qt GPL Exception version 1.3, included in
** the file GPL_EXCEPTION.txt in this package.
**
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
**
****************************************************************************/


#ifndef QSTATEMACHINE_P_H
#define QSTATEMACHINE_P_H

//
//  W A R N I N G
//  -------------
//
// This file is not part of the Qt API.  It exists purely as an
// implementation detail.  This header file may change from version to
// version without notice, or even be removed.
//
// We mean it.
//

#ifndef QT_STATEMACHINE_SOLUTION
#include <private/qobject_p.h>
#endif
#include <QtCore/qcoreevent.h>
#include <QtCore/qhash.h>
#include <QtCore/qlist.h>
#include <QtCore/qset.h>

#include "qstate.h"

QT_BEGIN_NAMESPACE

class QEvent;
#ifndef QT_NO_STATEMACHINE_EVENTFILTER
class QtEventTransition;
#endif
class QtSignalEventGenerator;
class QtSignalTransition;
class QtAbstractState;
class QtAbstractStateGroup;
class QtAbstractTransition;
class QtState;

#ifndef QT_NO_ANIMATION
class QtPropertyAnimation;
#endif

class QtStateMachine;
class Q_ANIMATION_EXPORT QtStateMachinePrivate
#ifndef QT_STATEMACHINE_SOLUTION
    : public QObjectPrivate
#endif
{
    Q_DECLARE_PUBLIC(QtStateMachine)
public:
    enum State {
        NotRunning,
        Starting,
        Running
    };
    enum StopProcessingReason {
        EventQueueEmpty,
        Finished,
        Stopped
    };

    QtStateMachinePrivate();
    ~QtStateMachinePrivate();

    static QtStateMachinePrivate *get(QtStateMachine *q);

    static QtAbstractStateGroup *findLCA(const QList<QtAbstractState*> &states);

    static bool stateEntryLessThan(QtAbstractState *s1, QtAbstractState *s2);
    static bool stateExitLessThan(QtAbstractState *s1, QtAbstractState *s2);

    QtAbstractState *findErrorState(QtAbstractState *context);
    void setError(QtStateMachine::Error error, QtAbstractState *currentContext);

    // private slots
    void _q_start();
    void _q_process();

    void microstep(const QList<QtAbstractTransition*> &transitionList);
    bool isPreempted(const QtAbstractState *s, const QSet<QtAbstractTransition*> &transitions) const;
    QSet<QtAbstractTransition*> selectTransitions(QEvent *event) const;
    void exitStates(const QList<QtAbstractTransition*> &transitionList);
    void executeTransitionContent(const QList<QtAbstractTransition*> &transitionList);
    void enterStates(const QList<QtAbstractTransition*> &enabledTransitions);
    void addStatesToEnter(QtAbstractState *s, QtAbstractStateGroup *root,
                          QSet<QtAbstractState*> &statesToEnter,
                          QSet<QtAbstractState*> &statesForDefaultEntry);

    bool isInFinalState(QtAbstractState *s) const;
    static bool isFinal(const QtAbstractState *s);
    static bool isParallel(const QtAbstractState *s);
    static bool isCompound(const QtAbstractState *s);
    static bool isAtomic(const QtAbstractState *s);
    static bool isDescendantOf(const QtAbstractState *s, const QtAbstractState *other);
    static QList<QtAbstractStateGroup*> properAncestors(const QtAbstractState *s, const QtAbstractStateGroup *upperBound);

    void registerTransitions(QtAbstractState *state);
    void registerSignalTransition(QtSignalTransition *transition);
#ifndef QT_NO_STATEMACHINE_EVENTFILTER
    void registerEventTransition(QtEventTransition *transition);
#endif
    void handleTransitionSignal(const QObject *sender, int signalIndex,
                                void **args);    
    void scheduleProcess();
    
    typedef QPair<QObject *, QByteArray> RestorableId;
    QHash<RestorableId, QVariant> registeredRestorables;
    void registerRestorable(QObject *object, const QByteArray &propertyName);
    void unregisterRestorable(QObject *object, const QByteArray &propertyName);
    bool hasRestorable(QObject *object, const QByteArray &propertyName) const;
    QVariant restorableValue(QObject *object, const QByteArray &propertyName) const;

#ifndef QT_NO_ANIMATION
    void registerRestorable(QtPropertyAnimation *animation);
        
    QtPropertyAnimation *restorableAnimation(QObject *object, const QByteArray &propertyName);
    QList<QtPropertyAnimation *> restorableAnimations() const;
    
    QHash<RestorableId, QtPropertyAnimation*> registeredRestorableAnimations;    
#endif // QT_NO_ANIMATION

    State state;
    bool processing;
    bool processingScheduled;
    bool stop;
    StopProcessingReason stopProcessingReason;
    QtState *rootState;
    QSet<QtAbstractState*> configuration;
    QList<QEvent*> internalEventQueue;
    QList<QEvent*> externalEventQueue;
    
    QtStateMachine::Error error;
    QtState::RestorePolicy globalRestorePolicy;

    QString errorString;
    QSet<QtAbstractState *> pendingErrorStates;
    QSet<QtAbstractState *> pendingErrorStatesForDefaultEntry;
    QtAbstractState *initialErrorStateForRoot;

#ifndef QT_STATEMACHINE_SOLUTION
    QtSignalEventGenerator *signalEventGenerator;
#endif
    QHash<const QObject*, QList<int> > connections;
#ifndef QT_NO_STATEMACHINE_EVENTFILTER
    QHash<QObject*, QSet<QEvent::Type> > qobjectEvents;
#endif
    QHash<int, QEvent*> delayedEvents;
  
    typedef QEvent* (*f_cloneEvent)(QEvent*);
    struct Handler {
        f_cloneEvent cloneEvent;
    };

    static const Handler *handler;

#ifdef QT_STATEMACHINE_SOLUTION
    QtStateMachine *q_ptr;
#endif
};

QT_END_NAMESPACE

#endif
